/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: muaddib_atreides@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdio.h>
#include <stdlib.h>

#include "nSX2.h"
#include "r5900_main.h"
#include "r5900_main_table.h"

_u8		rt;
_u8		rs;
_u8		rd;
_u8		base;
_u8		sa;
_u64	imm;
_u64	offset;
_u64	instr_index;
_u8		sel;
_u8		ft;


void PrintDebugOpcode() {
	printf("==>%8lx\n",MipsCpu.opcode);
	
	switch ( ((MipsCpu.opcode >> 26) & 0x3F) ) {
	
		case 0 	: printf("r5900:dbg_opcode_special: %s \n", debug_special_opc[(MipsCpu.opcode  & 0x3F)]); break;
		
		case 1	: printf("r5900:dbg_opcode_regimm: %s \n", debug_regimm_opc[(MipsCpu.opcode  & 0x1F0000)>>16]); break;
		
		case 16 : printf("r5900:dbg_opcode_cop0: %s \n", debug_cop0_opc[(MipsCpu.opcode&0x3E00000)>>21]);
				  printf("r5900:dbg_opcode_cop0c0: %s \n", debug_cop0c0_opc[(MipsCpu.opcode&0x3F)]);
							break;
		case 17 : {
					switch( (MipsCpu.opcode&0x3E00000)>>21) {
						
						case 20: printf("r5900:debug_cop1_W_opc: %s \n", debug_cop1_W_L_opc[MipsCpu.opcode&0x3F] );break;					
	
		
						default: printf("r5900:dbg_opcode_cop1: %s \n", debug_cop1_opc[(MipsCpu.opcode&0x3E00000)>>21]);break;					
					}
					
					break;
				}	
	
		case 28 : printf("r5900:dbg_opcode_special2: %s \n", debug_special2_opc[(MipsCpu.opcode  & 0x3F)] );
	
		
		default : printf("r5900:dbg_opcode: %s \n", debug_normal_opc[((MipsCpu.opcode >> 26) & 0x3F)]); break;
	
	}

}


void MipsUpdateCpu ( ) {
	
	MipsFillOpcode(MipsCpu.PC);
	
	if ( debug_op == 1 ) 	printf("%8llx ==> ", MipsCpu.PC );
	if ( MipsCpu.opcode != 0 ) 	{
		normal_opc[((MipsCpu.opcode >> 26) & 0x3F)]();
	}
	else {
			if ( debug_op == 1 ) 	printf("	NOP\n");
			MipsCpu.PC = MipsCpu.PC + 4;
				 	
			}
	
}

void MipsCpuPrintReg( void ) {
	_u8	i;

	for ( i = 0 ; i < 11; i++ ) {
	
	printf("%s : %16.16llx  | %s : %16.16llx | %s :%16.16llx  | \n",  
	debug_r5900_reg[ (i*3)+0 ], MipsCpu.GPR[(i*3)+0].lo ,debug_r5900_reg[(i*3)+1], MipsCpu.GPR[(i*3)+1].lo ,  
	debug_r5900_reg[(i*3)+2], MipsCpu.GPR[(i*3)+2].lo   );

	}
	
	printf("%s : %16.16llx  |      \n",  
	debug_r5900_reg[CPU_REG_LO], MipsCpu.GPR[CPU_REG_LO].lo 	  );

}


void SPECIAL() {	special_opc[(MipsCpu.opcode  & 0x3F)]();	}


void REGIMM() {		regimm_opc[(MipsCpu.opcode  & 0x1F0000)>>16]();		}


void J() {
	_u64 	target_branch;

	Extract_instr_index();
	
	target_branch = instr_index  << 2;

	if ( debug_op == 1 ) 	printf("j %8llx\n", target_branch );
	
	MipsCpu.PC = MipsCpu.PC + 4;
	MipsUpdateCpu();
	
	 MipsCpu.PC = ( MipsCpu.PC & 0xFFFFFFFF00000000 ) | target_branch;
}


void     JAL  () {

	_u64 	target_branch;

	Extract_instr_index();
	MipsCpu.GPR[31].lo =  MipsCpu.PC + 8;
	
	target_branch = instr_index  << 2;

	if ( debug_op == 1 ) 	printf("jal %8llx\n", target_branch );
	
	MipsCpu.PC = MipsCpu.PC + 4;
	MipsUpdateCpu();
	
	 MipsCpu.PC = ( MipsCpu.PC & 0xFFFFFFFF00000000 ) | target_branch;
}


void     BEQ () {
	_u64	target_branch;
	
	Extract_rs();
	Extract_rt();
	Extract_offset();
	
	offset = offset << 2;
	sign_extendedh(offset);
	target_branch = MipsCpu.PC + 4 +  offset ;
	if ( debug_op == 1 ) 	printf("beq	%s, %s,  %4llx\n", debug_r5900_reg[rs],debug_r5900_reg[rt],target_branch );

	if ( MipsCpu.GPR[rs].lo == MipsCpu.GPR[rt].lo ) {
			MipsCpu.PC = MipsCpu.PC + 4;
			MipsUpdateCpu();
			MipsCpu.PC = target_branch;
	}
	else MipsCpu.PC = MipsCpu.PC + 4;
} 


void     BNE () {

	_u64	target_branch;
	
	Extract_rs();
	Extract_rt();
	Extract_offset();
	
	offset = offset << 2;
	sign_extendedh(offset);
	target_branch = MipsCpu.PC + 4 +  offset ;
	if ( debug_op == 1 ) 	printf("bne	%s, %s,  %4llx\n", debug_r5900_reg[rs],debug_r5900_reg[rt],target_branch );

	if ( MipsCpu.GPR[rs].lo != MipsCpu.GPR[rt].lo ) {
			MipsCpu.PC = MipsCpu.PC + 4;
			MipsUpdateCpu();
			MipsCpu.PC = target_branch;
	}
	else MipsCpu.PC = MipsCpu.PC + 4;
	

}
void     BLEZ () {
	_u64	target_branch;
	
	Extract_rs();
	Extract_offset();
	
	offset = offset << 2;
	sign_extendedh(offset);
	target_branch = MipsCpu.PC + 4 +  offset ;
	if ( debug_op == 1 ) 	printf("blez	%s, %4llx\n", debug_r5900_reg[rs], target_branch );

	if ( (long long)(MipsCpu.GPR[rs].lo) <= 0 ) {
			MipsCpu.PC = MipsCpu.PC + 4;
			MipsUpdateCpu();
			MipsCpu.PC = target_branch;
	}
	else MipsCpu.PC = MipsCpu.PC + 4;
}

void     BGTZ () {
	_u64	target_branch;
	
	Extract_rs();
	Extract_offset();
	
	offset = offset << 2;
	sign_extendedh(offset);
	target_branch = MipsCpu.PC + 4 +  offset ;
	if ( debug_op == 1 ) 	printf("bgtz	%s, %4llx\n", debug_r5900_reg[rs], target_branch );

	

	if ( (long long)( MipsCpu.GPR[rs].lo ) > 0 ) {
			MipsCpu.PC = MipsCpu.PC + 4;
			MipsUpdateCpu();
			MipsCpu.PC = target_branch;
	}
	else MipsCpu.PC = MipsCpu.PC + 4;
}


void    ADDI   () {
	_u64 temp;
	
	Extract_rt(); 
	Extract_rs();
	Extract_imm();
	if ( debug_op == 1 ) 	printf("addi	%s, %s,  %4llx\n", debug_r5900_reg[rt],debug_r5900_reg[rs],imm );
	
	sign_extendedh(imm);
	temp = MipsCpu.GPR[rs].lo + imm;
	
//	if (( temp >> 32) == 0 ) {
	
		sign_extended(temp);
		MipsCpu.GPR[rt].lo = temp;
//	}
//	else
//	{
		//INTEGER OVERFLOW
//	}
	//printf(" %s=%8llx\n",debug_r5900_reg[rt],MipsCpu.GPR[rt].lo);	
	MipsCpu.PC = MipsCpu.PC + 4;
}


void    ADDIU () {
	_u64 temp;
	
	Extract_rt(); 
	Extract_rs();
	Extract_imm();
	if ( debug_op == 1 ) 	printf("addiu	%s, %s,  %4llx\n", debug_r5900_reg[rt],debug_r5900_reg[rs],imm );

	
	sign_extendedh(imm);
	temp = MipsCpu.GPR[rs].lo + imm;
	sign_extended(temp);
	MipsCpu.GPR[rt].lo = temp;


	MipsCpu.PC = MipsCpu.PC + 4;
}


void    SLTI() {
	
	Extract_rt(); 
	Extract_rs();
	Extract_imm();

	if ( debug_op == 1 ) 	printf("slti	%s, %s,  %4llx\n", debug_r5900_reg[rt],debug_r5900_reg[rs],imm );
	sign_extendedh(imm);
	
	if ( MipsCpu.GPR[rs].lo <  imm ) 	MipsCpu.GPR[rt].lo = 1;
	MipsCpu.GPR[rt].lo = 0;


	MipsCpu.PC = MipsCpu.PC + 4;
}


void    SLTIU() {
	Extract_rs();
	Extract_rt();
	Extract_imm();
	
	if ( debug_op == 1 ) 	printf("sltiu		%s, %s, %8llx\n", debug_r5900_reg[rd], debug_r5900_reg[rs], imm);
	sign_extendedh(imm);
 
    	if  ( 	MipsCpu.GPR[rs].lo < imm ) 	MipsCpu.GPR[rt].lo = 1; 
    	else 	MipsCpu.GPR[rt].lo = 0; 
	
	MipsCpu.PC = MipsCpu.PC + 4;
}


void    ANDI() {

	
	Extract_rt(); 
	Extract_rs();
	Extract_imm();
	if ( debug_op == 1 ) 	printf("andi	%s, %s,  %4llx\n", debug_r5900_reg[rt],debug_r5900_reg[rs],imm );

	MipsCpu.GPR[rt].lo = MipsCpu.GPR[rs].lo & imm;


	MipsCpu.PC = MipsCpu.PC + 4;
}


void    ORI () {

	
	Extract_rt(); 
	Extract_rs();
	Extract_imm();
	if ( debug_op == 1 ) 	printf("ori	%s, %s,  %4llx\n", debug_r5900_reg[rt],debug_r5900_reg[rs],imm );

	MipsCpu.GPR[rt].lo = MipsCpu.GPR[rs].lo | imm;


	MipsCpu.PC = MipsCpu.PC + 4;
}


void    XORI () {
	Extract_rt(); 
	Extract_rs();
	Extract_imm();
	if ( debug_op == 1 ) 	printf("xori	%s, %s,  %4llx\n", debug_r5900_reg[rt],debug_r5900_reg[rs],imm );

	MipsCpu.GPR[rt].lo = MipsCpu.GPR[rs].lo ^ imm;


	MipsCpu.PC = MipsCpu.PC + 4;
}


void    LUI  () {
	Extract_rt(); 
	Extract_imm();
	if ( debug_op == 1 ) 	printf("lui	%s, %4llx\n", debug_r5900_reg[rt],imm );
	
	imm = imm << 16;
	sign_extended(imm);
	MipsCpu.GPR[rt].lo = imm ;
	MipsCpu.PC = MipsCpu.PC + 4;
	

}


void    COP0   () {	cop0_opc[(MipsCpu.opcode&0x3E00000)>>21]();}


void    COP1  () {	cop1_opc[(MipsCpu.opcode&0x3E00000)>>21]();}


void    COP2  () {	MipsCpu.halt = 1;}
void    COP1X   () {	MipsCpu.halt = 1;}

void     BEQL() {
	_u64	target_branch;
	
	Extract_rs();
	Extract_rt();
	Extract_offset();
	
	offset = offset << 2;
	sign_extendedh(offset);
	target_branch = MipsCpu.PC + 4 +  offset ;
	if ( debug_op == 1 ) 	printf("beql	%s, %s,  %4llx\n", debug_r5900_reg[rs],debug_r5900_reg[rt],target_branch );

	if ( MipsCpu.GPR[rs].lo == MipsCpu.GPR[rt].lo ) {
			MipsCpu.PC = MipsCpu.PC + 4;
			MipsUpdateCpu();
			MipsCpu.PC = target_branch;
	}
	else MipsCpu.PC = MipsCpu.PC + 8;
}

void     BNEL() {
	_u64	target_branch;
	
	Extract_rs();
	Extract_rt();
	Extract_offset();
	
	offset = offset << 2;
	sign_extendedh(offset);
	target_branch = MipsCpu.PC + 4 +  offset ;
	if ( debug_op == 1 ) 	printf("bnel	%s, %s,  %4llx\n", debug_r5900_reg[rs],debug_r5900_reg[rt],target_branch );

	if ( MipsCpu.GPR[rs].lo != MipsCpu.GPR[rt].lo ) {
			MipsCpu.PC = MipsCpu.PC + 4;
			MipsUpdateCpu();
			MipsCpu.PC = target_branch;
	}
	else MipsCpu.PC = MipsCpu.PC + 8;
}


void    BLEZL() {	MipsCpu.halt = 1;}
void    BGTZL() {	MipsCpu.halt = 1;}
void    DADDI() {	MipsCpu.halt = 1;}

void    DADDIU    () {
	
	Extract_rt(); 
	Extract_rs();
	Extract_imm();
	if ( debug_op == 1 ) 	printf("daddiu	%s, %s,  %4llx\n", debug_r5900_reg[rt],debug_r5900_reg[rs],imm );
	
	sign_extendedh(imm);
	MipsCpu.GPR[rt].lo = MipsCpu.GPR[rs].lo + imm;

	MipsCpu.PC = MipsCpu.PC + 4;
}


void    LDL  () {
	Extract_base();
	Extract_rt();
	Extract_offset();
	
	if ( debug_op == 1 )		printf("ldl	%s, %4llx(%s)\n", debug_r5900_reg[rt],offset,debug_r5900_reg[base] );

	sign_extendedh(offset);
	
		
	MipsMemoryRead ( &MipsCpu.GPR[rt] , DATA_DWH ,  MipsCpu.GPR[base].lo , offset ) ;

	MipsCpu.PC = MipsCpu.PC + 4;
}


void    LDR   () {

	Extract_base();
	Extract_rt();
	Extract_offset();
	
	if ( debug_op == 1 )		printf("ldr	%s, %4llx(%s)\n", debug_r5900_reg[rt],offset,debug_r5900_reg[base] );

	sign_extendedh(offset);
	
		
	MipsMemoryRead ( &MipsCpu.GPR[rt] , DATA_DW ,  MipsCpu.GPR[base].lo , offset ) ;

	MipsCpu.PC = MipsCpu.PC + 4;
	

}


void    SPECIAL2  () {

//	printf("r5900:dbg_opcode_special2: %s (%x ) \n", debug_special2_opc[(MipsCpu.opcode  & 0x3F)], (MipsCpu.opcode  & 0x3F));
	special2_opc[(MipsCpu.opcode  & 0x3F)]();
}

 
void    JALX  () {	MipsCpu.halt = 1;}
void    MDMX   () {	MipsCpu.halt = 1;}
void    NI   () {	MipsCpu.halt = 1;}

void    LB     () {
	
	Extract_base();
	Extract_rt();
	Extract_offset();
	
	if ( debug_op == 1 ) 		printf("lb	%s, %4llx(%s)\n", debug_r5900_reg[rt],offset,debug_r5900_reg[base] );

	sign_extendedh(offset);
	
		
	MipsMemoryRead ( &MipsCpu.GPR[rt] , DATA_B ,  MipsCpu.GPR[base].lo , offset ) ;

	sign_extendedb( MipsCpu.GPR[rt].lo );

	MipsCpu.PC = MipsCpu.PC + 4;
}


void    LH    () {
	Extract_base();
	Extract_rt();
	Extract_offset();
	
	if ( debug_op == 1 ) 		printf("lh	%s, %4llx(%s)\n", debug_r5900_reg[rt],offset,debug_r5900_reg[base] );

	sign_extendedh(offset);
	MipsMemoryRead ( &MipsCpu.GPR[rt] , DATA_HW ,  MipsCpu.GPR[base].lo , offset ) ;
	sign_extendedh( MipsCpu.GPR[rt].lo );

	MipsCpu.PC = MipsCpu.PC + 4;
}

void    LWL () { MipsCpu.halt = 1; }


void    LW   () {
	Extract_base();
	Extract_rt();
	Extract_offset();
	
	if ( debug_op == 1 ) 		printf("lw	%s, %4llx(%s)\n", debug_r5900_reg[rt],offset,debug_r5900_reg[base] );

	sign_extendedh(offset);
	MipsMemoryRead ( &MipsCpu.GPR[rt] , DATA_W ,  MipsCpu.GPR[base].lo , offset ) ;
	sign_extended( MipsCpu.GPR[rt].lo );

	MipsCpu.PC = MipsCpu.PC + 4;
}


void    LBU () {
	Extract_base();
	Extract_rt();
	Extract_offset();
	
	if ( debug_op == 1 ) 		printf("lbu	%s, %4llx(%s)\n", debug_r5900_reg[rt],offset,debug_r5900_reg[base] );

	sign_extendedh(offset);
	MipsMemoryRead ( &MipsCpu.GPR[rt] , DATA_B ,  MipsCpu.GPR[base].lo , offset ) ;

	MipsCpu.PC = MipsCpu.PC + 4;
}


void    LHU () {
	Extract_base();
	Extract_rt();
	Extract_offset();
	
	if ( debug_op == 1 ) 		printf("lhu	%s, %4llx(%s)\n", debug_r5900_reg[rt],offset,debug_r5900_reg[base] );

	sign_extendedh(offset);
	MipsMemoryRead ( &MipsCpu.GPR[rt] , DATA_HW ,  MipsCpu.GPR[base].lo , offset ) ;

	MipsCpu.PC = MipsCpu.PC + 4;
}


void    LWR  () {	MipsCpu.halt = 1;}


void    LWU   () {
	Extract_base();
	Extract_rt();
	Extract_offset();
	
	if ( debug_op == 1 ) 		printf("lwu	%s, %4llx(%s)\n", debug_r5900_reg[rt],offset,debug_r5900_reg[base] );

	sign_extendedh(offset);
	MipsMemoryRead ( &MipsCpu.GPR[rt] , DATA_W ,  MipsCpu.GPR[base].lo , offset ) ;

	MipsCpu.PC = MipsCpu.PC + 4;
}

void    SB     () {
	Extract_base();
	Extract_rt();
	Extract_offset();
	
	if ( debug_op == 1 ) 	printf("sb		%s, %4llx(%s)\n", debug_r5900_reg[rt],offset,debug_r5900_reg[base] );

	sign_extendedh(offset);
	MipsMemoryWrite ( MipsCpu.GPR[rt] , DATA_B ,  MipsCpu.GPR[base].lo , offset ) ;

	MipsCpu.PC = MipsCpu.PC + 4;
}

void    SH    () {
	Extract_base();
	Extract_rt();
	Extract_offset();
	
	if ( debug_op == 1 ) 	printf("sh		%s, %4llx(%s)\n", debug_r5900_reg[rt],offset,debug_r5900_reg[base] );

	sign_extendedh(offset);
	MipsMemoryWrite ( MipsCpu.GPR[rt] , DATA_HW ,  MipsCpu.GPR[base].lo , offset ) ;

	MipsCpu.PC = MipsCpu.PC + 4;
}


void    SWL () {	MipsCpu.halt = 1;}


void    SW   () {
	Extract_base();
	Extract_rt();
	Extract_offset();
	
	if ( debug_op == 1 ) 	printf("sd		%s, %4llx(%s)\n", debug_r5900_reg[rt],offset,debug_r5900_reg[base] );
	
	sign_extendedh(offset);
	MipsMemoryWrite ( MipsCpu.GPR[rt] , DATA_W ,  MipsCpu.GPR[base].lo , offset ) ;

	MipsCpu.PC = MipsCpu.PC + 4;
}


void    SDL  () {
	Extract_base();
	Extract_rt();
	Extract_offset();
	
	if ( debug_op == 1 ) 	printf("sdl	%s, %4llx(%s)\n", debug_r5900_reg[rt],offset,debug_r5900_reg[base] );

	sign_extendedh(offset);
	MipsMemoryWrite ( MipsCpu.GPR[rt] , DATA_DWH ,  MipsCpu.GPR[base].lo , offset ) ;

	MipsCpu.PC = MipsCpu.PC + 4;
}


void    SDR  () {
	Extract_base();
	Extract_rt();
	Extract_offset();
	
	if ( debug_op == 1 ) 	printf("	sd	r	%s, %4llx(%s)\n", debug_r5900_reg[rt],offset,debug_r5900_reg[base] );

	sign_extendedh(offset);
	MipsMemoryWrite ( MipsCpu.GPR[rt] , DATA_DW ,  MipsCpu.GPR[base].lo , offset ) ;

	MipsCpu.PC = MipsCpu.PC + 4;
}


void    SWR  () {	MipsCpu.halt = 1;}

void    CACHE( ) {

	printf("CACHE\n");
	MipsCpu.PC = 	MipsCpu.PC + 4 ;
}


void    LL   () {	MipsCpu.halt = 1;}


void    LWC1 () {	
	Extract_base();
	Extract_rt();
	Extract_offset();
	
	if ( debug_op == 1 ) 		printf("lwc1	FPR1_%2.2x, %4llx(%s)\n", rt,offset,debug_r5900_reg[base] );
	
	sign_extendedh(offset);
	MipsMemoryRead ( &MipsCpu.FPR1[ft] , DATA_W ,  MipsCpu.GPR[base].lo , offset ) ;
	//sign_extended( MipsCpu.GPR[rt].lo );

	MipsCpu.PC = MipsCpu.PC + 4;
}


void    LWC2 () {	MipsCpu.halt = 1;}
void    PREF () {	MipsCpu.halt = 1;}
void    LLD  () {	MipsCpu.halt = 1;}
void    LDC1 () {	MipsCpu.halt = 1;}
void    LDC2 () {	MipsCpu.halt = 1;}


void    LD   () {
	Extract_base();
	Extract_rt();
	Extract_offset();
	
	if ( debug_op == 1 ) 		printf("ld	%s, %4llx(%s)\n", debug_r5900_reg[rt],offset,debug_r5900_reg[base] );
	
	sign_extendedh(offset);
	MipsMemoryRead ( &MipsCpu.GPR[rt] , DATA_DW ,  MipsCpu.GPR[base].lo , offset ) ;
	//sign_extended( MipsCpu.GPR[rt].lo );

	MipsCpu.PC = MipsCpu.PC + 4;
	

}

void    SC     () {	MipsCpu.halt = 1;}

void    SWC1    () {

	_u128	temp;	
	
	Extract_base();
	Extract_ft();
	Extract_offset();


	
	if ( debug_op == 1 ) 	printf("sdwc1	fp%d, %4llx(%s)\n", ft,offset,debug_r5900_reg[base] );

	sign_extendedh(offset);
	//temp.lo =  MipsCpu.FPR1[ft];
	MipsMemoryWrite ( MipsCpu.FPR1[ft] , DATA_W ,  MipsCpu.GPR[base].lo , offset ) ;

	MipsCpu.PC = MipsCpu.PC + 4;
}


void    SWC2  () {	MipsCpu.halt = 1;}


void SQ() {

	Extract_base();
	Extract_rt();
	Extract_offset();
	
	if ( debug_op == 1 ) 	printf("sq		%s, %4llx(%s)\n", debug_r5900_reg[rt],offset,debug_r5900_reg[base] );

	sign_extendedh(offset);
	MipsMemoryWrite ( MipsCpu.GPR[rt] , DATA_QUAD ,  MipsCpu.GPR[base].lo , offset ) ;
	MipsCpu.PC = MipsCpu.PC + 4;
}

void    SCD  () {	MipsCpu.halt = 1;}
void    SDC1 () {	MipsCpu.halt = 1;}
void    SDC2 () {	MipsCpu.halt = 1;}


void    SD() {
	Extract_base();
	Extract_rt();
	Extract_offset();
	
	if ( debug_op == 1 ) 	printf("sd		%s, %4llx(%s)\n", debug_r5900_reg[rt],offset,debug_r5900_reg[base] );

	sign_extendedh(offset);
	MipsMemoryWrite ( MipsCpu.GPR[rt] , DATA_DW ,  MipsCpu.GPR[base].lo , offset ) ;

	MipsCpu.PC = MipsCpu.PC + 4;
}


void   SLL() {
	_u64 temp;
	
	Extract_rt();
	Extract_rd();
	Extract_sa();
	
	if ( debug_op == 1 ) 	printf("sll		%s, %s , %d \n", debug_r5900_reg[rd], debug_r5900_reg[rt], sa );
	
	temp = MipsCpu.GPR[rt].lo <<  sa  ;	
	sign_extended(temp);
	MipsCpu.GPR[rd].lo = temp;
	
	MipsCpu.PC = MipsCpu.PC + 4;
}


void   MOVCI() {	MipsCpu.halt = 1;}


void   SRL   () {
	_u64 temp;
	
	Extract_rt();
	Extract_rd();
	Extract_sa();
	
	if ( debug_op == 1 ) 	printf("srl		%s, %s , %d\n", debug_r5900_reg[rd], debug_r5900_reg[rt], sa );
	
	temp = MipsCpu.GPR[rt].lo  >>  sa  ;	
	sign_extended(temp);
	MipsCpu.GPR[rd].lo = temp;
	
	MipsCpu.PC = MipsCpu.PC + 4;
}


void   SRA  () {
	_u64 temp;

	
	Extract_rt();
	Extract_rd();
	Extract_sa();
	
	if ( debug_op == 1 ) 	printf("sra		%s, %s , %d\n", debug_r5900_reg[rd], debug_r5900_reg[rt], sa );

	
	if (MipsCpu.GPR[rt].lo & 0x0000000080000000LL) {
		
		
		
		
		printf("NEED SRA ADJUST\n");
	} else { temp =  MipsCpu.GPR[rt].lo >>  sa  ; }
	
	sign_extended(temp);
	MipsCpu.GPR[rd].lo = temp;
	
	MipsCpu.PC = MipsCpu.PC + 4;
	


}
 
void   SLLV () {	MipsCpu.halt = 1;}
void   SLRV () {	MipsCpu.halt = 1;}
void   SRAV () {	MipsCpu.halt = 1;}


void   JR   () {
	_u64	branch_target;
	Extract_rs();

	if ( debug_op == 1 ) 	printf("	jr	%s with %s = %8llx\n", debug_r5900_reg[rs],debug_r5900_reg[rs] , MipsCpu.GPR[rs].lo );

	branch_target = MipsCpu.GPR[rs].lo;
	MipsCpu.PC = MipsCpu.PC + 4;
	MipsUpdateCpu();
	MipsCpu.PC =  branch_target;
}


void   JALR () {
	_u64	branch_target;
	Extract_rs();
	Extract_rd();
	
	branch_target = MipsCpu.GPR[rs].lo;
	MipsCpu.GPR[rd].lo = MipsCpu.PC + 8;
	
	if ( debug_op == 1 ) 	printf("jalr	%s, %s with  %s = %8llx\n", debug_r5900_reg[rd],debug_r5900_reg[rs] ,debug_r5900_reg[rs], MipsCpu.GPR[rs].lo );

	MipsCpu.PC = MipsCpu.PC + 4;
	MipsUpdateCpu();

	MipsCpu.PC =  branch_target;
}

void   MOVZ() {	MipsCpu.halt = 1;}


void   MOVN() {
	Extract_rs();
	Extract_rt();
	Extract_rd();
	
	if ( debug_op == 1 ) 	printf("movn	%s, %s %s \n", debug_r5900_reg[rd],debug_r5900_reg[rs] ,debug_r5900_reg[rt] );
	
	if(  MipsCpu.GPR[rt].lo != 0 ) MipsCpu.GPR[rd].lo = MipsCpu.GPR[rs].lo ;
		
	MipsCpu.PC = MipsCpu.PC + 4;
	
	
}

void   SYSCALL() {

	if ( debug_print_syscall == 1 )	printf("%8llx : %s ( %x h) \n", MipsCpu.PC , debug_syscall[(MipsCpu.GPR[3].lo & 0xFF)],(_u8) (MipsCpu.GPR[3].lo & 0xFF));
/*
	printf("	v0: %16.16llx\n", MipsCpu.GPR[2].lo );
	printf("	v1: %16.16llx\n", MipsCpu.GPR[3].lo );	
	printf("	a0: %16.16llx\n", MipsCpu.GPR[4].lo );
	printf("	a1: %16.16llx\n", MipsCpu.GPR[5].lo );	
	printf("	a2: %16.16llx\n", MipsCpu.GPR[6].lo );	
	getchar();
*/
	syscall_fct[(MipsCpu.GPR[3].lo & 0xFF)]();

}

void   BREAK () {	MipsCpu.halt = 1;}


void   SYNC  () {
	printf("SYNC\n");
	MipsCpu.PC = MipsCpu.PC + 4;
}


void   MFHI   () {
	Extract_rd();
	
	MipsCpu.GPR[rd].lo = MipsCpu.GPR[CPU_REG_HI].lo;

	MipsCpu.PC = MipsCpu.PC + 4;
}

void   MTHI  () {	MipsCpu.halt = 1;}
void   MFLO  () {	MipsCpu.halt = 1;}
void   MTLO  () {	MipsCpu.halt = 1;}
void   DSLLV () {	MipsCpu.halt = 1;}
void   DSRLV () {	MipsCpu.halt = 1;}
void   DSRAV () {	MipsCpu.halt = 1;}

void   MULT     () {
	Extract_rs();
	Extract_rt();
	Extract_rd();			

	if ( debug_op == 1 ) printf("mult	%s, %s, %s \n", debug_r5900_reg[rd] , debug_r5900_reg[rs],debug_r5900_reg[rt] );

	MipsCpu.GPR[CPU_REG_LO].lo = ( MipsCpu.GPR[rs].lo & 0xFFFFFFFF ) * ( MipsCpu.GPR[rt].lo & 0xFFFFFFFF );
	MipsCpu.GPR[CPU_REG_HI].lo = ( (MipsCpu.GPR[rs].lo >> 32 ) & 0xFFFFFFFF ) * ( (MipsCpu.GPR[rt].lo >> 32 )& 0xFFFFFFFF );
	
	sign_extended(MipsCpu.GPR[CPU_REG_LO].lo);
	sign_extended(MipsCpu.GPR[CPU_REG_HI].lo);
	
	MipsCpu.GPR[rd].lo = MipsCpu.GPR[CPU_REG_LO].lo ;
	
	MipsCpu.PC = MipsCpu.PC + 4;
}

void   MULTU() {	MipsCpu.halt = 1;}
void   DIV  () {	MipsCpu.halt = 1;}

void   DIVU () {

	
	div_t result_div;
	Extract_rs();
	Extract_rt();
	
	if ( debug_op == 1 )  	printf("	divu	%s, %s\n", debug_r5900_reg[rs],debug_r5900_reg[rt]);

	
	result_div = div(  (MipsCpu.GPR[rs].lo & 0xFFFFFFFF )  ,  (MipsCpu.GPR[rt].lo & 0xFFFFFFFF )  );
	MipsCpu.GPR[CPU_REG_LO].lo = result_div.quot;
	MipsCpu.GPR[CPU_REG_HI].lo = result_div.rem;

	sign_extended( MipsCpu.GPR[CPU_REG_LO].lo);
	sign_extended( MipsCpu.GPR[CPU_REG_HI].lo);

	MipsCpu.PC = MipsCpu.PC + 4;
}

void   DMULT  () {	MipsCpu.halt = 1;}
void   DMULTU () {	MipsCpu.halt = 1;}
void   DDIV   () {	MipsCpu.halt = 1;}
void   DDIVU  () {	MipsCpu.halt = 1;}

void   ADD     () {
	_u64 temp;
	_u64 check;
	
	Extract_rt(); 
	Extract_rs();
	Extract_rd();
	
	if ( debug_op == 1 ) 	printf("add	%s, %s,  %s\n", debug_r5900_reg[rd],debug_r5900_reg[rt],debug_r5900_reg[rs] );
	
	
	temp = MipsCpu.GPR[rs].lo + MipsCpu.GPR[rt].lo;
	
	check =  ( temp >> 30 ) & 3;
	
	if ( (check == 0) ||(check == 3) ) {
		sign_extended(temp);
		MipsCpu.GPR[rd].lo = temp;
	} else {
		printf("add Integer Overflow\n" );
	}
	

	MipsCpu.PC = MipsCpu.PC + 4;
}

void   ADDU    () {
	_u64 temp;
	
	Extract_rt(); 
	Extract_rs();
	Extract_rd();
	if ( debug_op == 1 ) 	printf("addu	%s, %s,  %s\n", debug_r5900_reg[rd],debug_r5900_reg[rt],debug_r5900_reg[rs] );
	
	
	temp = MipsCpu.GPR[rs].lo + MipsCpu.GPR[rt].lo;
	sign_extended(temp);
	MipsCpu.GPR[rd].lo = temp;

	MipsCpu.PC = MipsCpu.PC + 4;

}

void   SUB () {
	_u64 temp;
	_u64 check;
	
	Extract_rt(); 
	Extract_rs();
	Extract_rd();
	if ( debug_op == 1 ) 	printf("sub	%s, %s,  %s\n", debug_r5900_reg[rd],debug_r5900_reg[rs],debug_r5900_reg[rt] );
	
	temp = MipsCpu.GPR[rs].lo - MipsCpu.GPR[rt].lo;	
	check =  ( temp >> 30 ) & 3;
	
	if ( (check == 0) ||(check == 3) ) {
		sign_extended(temp);
		MipsCpu.GPR[rd].lo = temp;
	} else {
		printf("sub Integer Overflow\n" );
	}
	

	MipsCpu.PC = MipsCpu.PC + 4;
}


void   SUBU   () {
	_u64 temp;
	
	Extract_rt(); 
	Extract_rs();
	Extract_rd();

	if ( debug_op == 1 ) 	printf("subu	%s, %s,  %s\n", debug_r5900_reg[rd],debug_r5900_reg[rs],debug_r5900_reg[rt] );
	
	
	temp = MipsCpu.GPR[rs].lo - MipsCpu.GPR[rt].lo;
	sign_extended(temp);
	MipsCpu.GPR[rd].lo = temp;

	MipsCpu.PC = MipsCpu.PC + 4;
}


void   AND () {
	Extract_rt(); 
	Extract_rs();
	Extract_rd();
	if ( debug_op == 1 ) 	printf("and	%s, %s,  %s\n", debug_r5900_reg[rd],debug_r5900_reg[rs],debug_r5900_reg[rt] );

	MipsCpu.GPR[rd].lo = MipsCpu.GPR[rs].lo & MipsCpu.GPR[rt].lo;


	MipsCpu.PC = MipsCpu.PC + 4;
}


void   OR () {
	Extract_rt(); 
	Extract_rs();
	Extract_rd();
	if ( debug_op == 1 ) 	printf("or	%s, %s,  %s\n", debug_r5900_reg[rd],debug_r5900_reg[rs],debug_r5900_reg[rt] );

	MipsCpu.GPR[rd].lo = MipsCpu.GPR[rs].lo | MipsCpu.GPR[rt].lo;

	MipsCpu.PC = MipsCpu.PC + 4;
}


void   XOR  () {
	Extract_rt(); 
	Extract_rs();
	Extract_rd();
	if ( debug_op == 1 ) 	printf("or	%s, %s,  %s\n", debug_r5900_reg[rd],debug_r5900_reg[rs],debug_r5900_reg[rt] );

	MipsCpu.GPR[rd].lo = MipsCpu.GPR[rs].lo ^ MipsCpu.GPR[rt].lo;

	MipsCpu.PC = MipsCpu.PC + 4;
}


void   NOR   () {	MipsCpu.halt = 1;}


void   SLT () {
	_u64	temp1;
	_u64	temp2;
	
	Extract_rs();
	Extract_rt();
	Extract_rd();
	
	if ( debug_op == 1 ) 	printf("slt		%s, %s, %s\n", debug_r5900_reg[rd], debug_r5900_reg[rs], debug_r5900_reg[rt]);
	
	temp1 = MipsCpu.GPR[rs].lo;
	temp2 = MipsCpu.GPR[rt].lo;
	
	sign_extended(temp1);
	sign_extended(temp2);

    if  ( temp1 < temp2 ) MipsCpu.GPR[rd].lo = 1; 
    else 	MipsCpu.GPR[rd].lo = 0; 

	MipsCpu.PC = MipsCpu.PC + 4;
}


void   SLTU   () {
	Extract_rs();
	Extract_rt();
	Extract_rd();
	
	if ( debug_op == 1 ) 	printf("sltu		%s, %s, %s\n", debug_r5900_reg[rd], debug_r5900_reg[rs], debug_r5900_reg[rt]);
 
    if  ( 	MipsCpu.GPR[rs].lo < 	MipsCpu.GPR[rt].lo ) 	MipsCpu.GPR[rd].lo = 1; 
    else 	MipsCpu.GPR[rd].lo = 0; 

	MipsCpu.PC = MipsCpu.PC + 4;
}

void   DADD  () {	MipsCpu.halt = 1;}


void   DADDU  () {
	
	_u64 temp;
	_u64 check;
	
	Extract_rs();
	Extract_rt();
	Extract_rd();
	
	if ( debug_op == 1 ) 	printf("daddu		%s, %s, %s\n", debug_r5900_reg[rd], debug_r5900_reg[rs], debug_r5900_reg[rt]);
	
	temp = MipsCpu.GPR[rs].lo + MipsCpu.GPR[rt].lo;
	check =  ( temp >> 62 ) & 3;
	if ( (check == 0) ||(check == 3) ) {
		sign_extended(temp);
		MipsCpu.GPR[rd].lo = temp;
	} else {
		printf("daddu Integer Overflow\n" );
	}
	
	MipsCpu.PC = MipsCpu.PC + 4;
}

void   DSUB  () {	MipsCpu.halt = 1;}

void   DSUBU() {
	_u64 temp;
	_u64 check;
	
	Extract_rt(); 
	Extract_rs();
	Extract_rd();
	if ( debug_op == 1 ) 	printf("dsubu	%s, %s,  %s\n", debug_r5900_reg[rd],debug_r5900_reg[rs],debug_r5900_reg[rt] );
	
	
	temp = MipsCpu.GPR[rs].lo - MipsCpu.GPR[rt].lo;
	check =  ( temp >> 62 ) & 3;
	if ( (check == 0) ||(check == 3) ) {
		sign_extended(temp);
		MipsCpu.GPR[rd].lo = temp;
	} else {
		printf("dsubu Integer Overflow\n" );
	}
	
	MipsCpu.PC = MipsCpu.PC + 4;

}

void   TGE   () {	MipsCpu.halt = 1;}
void   TGEU  () {	MipsCpu.halt = 1;}
void   TLT   () {	MipsCpu.halt = 1;}
void   TLTU  () {	MipsCpu.halt = 1;}
void   TEQ   () {	MipsCpu.halt = 1;}
void   TNE   () {	MipsCpu.halt = 1;}

void   DSLL     () {
	Extract_rt();
	Extract_rd();
	Extract_sa();
	
	if ( debug_op == 1 ) 	printf("dsll	%s, %s %4x\n", debug_r5900_reg[rd], debug_r5900_reg[rt],sa );
	
	MipsCpu.GPR[rd].lo = MipsCpu.GPR[rt].lo << sa ;

	MipsCpu.PC = MipsCpu.PC + 4;
}

void   DSRL  () {
	Extract_rt();
	Extract_rd();
	Extract_sa();
	
	if ( debug_op == 1 ) 	printf("dsrl	%s, %s %4x\n", debug_r5900_reg[rd], debug_r5900_reg[rt],sa );
	
	MipsCpu.GPR[rd].lo = MipsCpu.GPR[rt].lo >> sa ;

	MipsCpu.PC = MipsCpu.PC + 4;
}

void   DSRA   () {	MipsCpu.halt = 1;}

void   DSLL32  () {
	Extract_rt();
	Extract_rd();
	Extract_sa();
	
	if ( debug_op == 1 ) 	printf("dsll32	%s, %s %4x\n", debug_r5900_reg[rd], debug_r5900_reg[rt],sa );
	
	MipsCpu.GPR[rd].lo = MipsCpu.GPR[rt].lo << ( sa + 32 ) ;

	MipsCpu.PC = MipsCpu.PC + 4;
}


void   DSRL32   () {
	Extract_rt();
	Extract_rd();
	Extract_sa();
	
	if ( debug_op == 1 ) 	printf("dsrl32	%s, %s %4x\n", debug_r5900_reg[rd], debug_r5900_reg[rt],sa );
	
	MipsCpu.GPR[rd].lo = MipsCpu.GPR[rt].lo >> ( sa + 32 ) ;
	
	MipsCpu.PC = MipsCpu.PC + 4;
}


void   DSRA32() {
	_u64 temp;
	
	Extract_rt();
	Extract_rd();
	Extract_sa();
	
	if ( debug_op == 1 ) 	printf("dsra32	%s, %s %4x\n", debug_r5900_reg[rd], debug_r5900_reg[rt],sa );

	if (MipsCpu.GPR[rt].lo & 0x80000000000000000LL) { 
		
		
		
		printf("NEED dsra32 Adjust\n");
	} else {
	
		temp = MipsCpu.GPR[rt].lo >> ( sa + 32 ) ;
		MipsCpu.GPR[rd].lo = temp;
	}
	
	MipsCpu.PC = MipsCpu.PC + 4;
}


void	BLTZ() {
	_u64	target_branch;
	
	Extract_rs();
	Extract_offset();
	
	offset = offset << 2;
	sign_extendedh(offset);
	target_branch = MipsCpu.PC + 4 +  offset ;
	if ( debug_op == 1 ) 	printf("bltz	%s, %4llx\n", debug_r5900_reg[rs], target_branch );

	if ((long long)( MipsCpu.GPR[rs].lo) < 0 ) {
			MipsCpu.PC = MipsCpu.PC + 4;
			MipsUpdateCpu();
			MipsCpu.PC = target_branch;
	}
	else MipsCpu.PC = MipsCpu.PC + 4;
}

void      BGEZ() {
	_u64	target_branch;
	
	Extract_rs();
	Extract_offset();
	
	offset = offset << 2;
	sign_extendedh(offset);
	target_branch = MipsCpu.PC + 4 +  offset ;
	if ( debug_op == 1 ) 	printf("bgez	%s, %4llx\n", debug_r5900_reg[rs], target_branch );

	if ( (long long)(MipsCpu.GPR[rs].lo) >= 0 ) {
			MipsCpu.PC = MipsCpu.PC + 4;
			MipsUpdateCpu();
			MipsCpu.PC = target_branch;
	}
	else MipsCpu.PC = MipsCpu.PC + 4;
}

void      BLTZL  () {	MipsCpu.halt = 1;}
void      BGEZL  () {	MipsCpu.halt = 1;}
void      TGEI   () {	MipsCpu.halt = 1;}
void      TGEIU  () {	MipsCpu.halt = 1;}
void      TLTI   () {	MipsCpu.halt = 1;}
void      TLTIU  () {	MipsCpu.halt = 1;}
void      TEQI   () {	MipsCpu.halt = 1;}
void      TNEI   () {	MipsCpu.halt = 1;}
void      BLTZAL () {	MipsCpu.halt = 1;}
void      BGEZAL () {	MipsCpu.halt = 1;}
void      BLTZALL() {	MipsCpu.halt = 1;}
void      BGEZALL() {	MipsCpu.halt = 1;}

void   	MFC0  () {
	_u64 data;

	Extract_rt();
	Extract_rd();
	Extract_sel();


	if ( debug_op == 1 ) printf("	mfc0	%s, %s,  %d\n", debug_r5900_reg[rt],debug_r5900_cop0_reg[rd],sel );
	
	data = MipsCpu.CPR0[rd][sel];
	sign_extended(data);
	
	MipsCpu.GPR[rt].lo = data ;
	MipsCpu.PC = MipsCpu.PC + 4;
	
}


void   	DMFC0 () {	MipsCpu.halt = 1;}
void   	MTC0  () {	MipsCpu.halt = 1;}
void   	DMTC0 () {	MipsCpu.halt = 1;}


void	C0    () {

	cop0c0_opc[(MipsCpu.opcode&0x3F)]();
	
}

void	TLBR  () {	MipsCpu.halt = 1;}
void	TLBWI () {	MipsCpu.halt = 1;}
void	TLBWR () {	MipsCpu.halt = 1;}
void	TLBP  () {	MipsCpu.halt = 1;}	
void	ERET  () {	MipsCpu.halt = 1;}
void	DERET () {	MipsCpu.halt = 1;}
void 	WAIT  () {	MipsCpu.halt = 1;}


void 	EI() {
	
	MipsCpu.CPR0[12][0]	=	MipsCpu.CPR0[12][0]  | 0x00010000;
	
	MipsCpu.PC = 	MipsCpu.PC + 4;
}


void 	DI() {
	MipsCpu.CPR0[12][0]	=	MipsCpu.CPR0[12][0]  & 0xFFFEFFFF;
	
	MipsCpu.PC = 	MipsCpu.PC + 4;
}

void SYSCALL_NI() {

	printf("Syscall_NI : %d ( %x h)\n",(char)MipsCpu.GPR[3].lo,(_u8)MipsCpu.GPR[3].lo);
	
	MipsCpu.halt = 1;
}


void SYSCALL_RFU000_FullReset( ) { MipsCpu.halt = 1; } 
void SYSCALL_ResetEE( ) { MipsCpu.halt = 1; } 

void SYSCALL_SetGsCrt( ) { 
	
	Ps2Bios.GsCrt = 1;
	MipsCpu.PC = 	MipsCpu.PC + 4;	
} 

void SYSCALL_RFU003( ) { MipsCpu.halt = 1; } 
void SYSCALL_Exit( ) { MipsCpu.halt = 1; } 
void SYSCALL_RFU005( ) { MipsCpu.halt = 1; } 
void SYSCALL_LoadPS2Exe( ) { MipsCpu.halt = 1; } 
void SYSCALL_ExecPS2( ) { MipsCpu.halt = 1; } 
void SYSCALL_RFU008( ) { MipsCpu.halt = 1; } 
void SYSCALL_RFU009( ) { MipsCpu.halt = 1; } 
void SYSCALL_AddSbusIntcHandler( ) { MipsCpu.halt = 1; } 
void SYSCALL_RemoveSbusIntcHandler( ) { MipsCpu.halt = 1; } 
void SYSCALL_Interrupt2Iop( ) { MipsCpu.halt = 1; } 
void SYSCALL_SetVTLBRefillHandler( ) { MipsCpu.halt = 1; } 
void SYSCALL_SetVCommonHandler( ) { MipsCpu.halt = 1; } 
void SYSCALL_SetVInterruptHandler( ) { MipsCpu.halt = 1; } 
void SYSCALL_AddIntcHandler( ) { MipsCpu.halt = 1; } 
void SYSCALL_RemoveIntcHandler( ) { MipsCpu.halt = 1; } 

void SYSCALL_AddDmacHandler( ) {
	
	MipsCpu.PC = MipsCpu.PC + 4;
 } 

void SYSCALL_RemoveDmacHandler( ) { MipsCpu.halt = 1; } 
void SYSCALL__EnableIntc( ) { MipsCpu.halt = 1; } 
void SYSCALL__DisableIntc( ) { MipsCpu.halt = 1; } 

void SYSCALL__EnableDmac( )  {
	
	MipsCpu.PC = MipsCpu.PC + 4;
 } 

void SYSCALL__DisableDmac( ) { MipsCpu.halt = 1; } 
void SYSCALL__SetAlarm( ) { MipsCpu.halt = 1; } 
void SYSCALL__ReleaseAlarm( ) { MipsCpu.halt = 1; } 
void SYSCALL_CreateThread( ) { MipsCpu.halt = 1; } 
void SYSCALL_DeleteThread( ) { MipsCpu.halt = 1; } 
void SYSCALL_StartThread( ) { MipsCpu.halt = 1; } 
void SYSCALL_ExitThread( ) { MipsCpu.halt = 1; } 
void SYSCALL_ExitDeleteThread( ) { MipsCpu.halt = 1; } 
void SYSCALL_TerminateThread( ) { MipsCpu.halt = 1; } 
void SYSCALL_DisableDispatchThread( ) { MipsCpu.halt = 1; } 
void SYSCALL_EnableDispatchThread( ) { MipsCpu.halt = 1; } 
void SYSCALL_ChangeThreadPriority( ) { MipsCpu.halt = 1; } 
void SYSCALL_RotateThreadReadyQueue( ) { MipsCpu.halt = 1; } 
void SYSCALL_ReleaseWaitThread( ) { MipsCpu.halt = 1; } 
void SYSCALL_GetThreadId( ) { MipsCpu.halt = 1; } 
void SYSCALL_ReferThreadStatus( ) { MipsCpu.halt = 1; } 
void SYSCALL_SleepThread( ) { MipsCpu.halt = 1; } 
void SYSCALL_WakeupThread( ) { MipsCpu.halt = 1; } 
void SYSCALL_CancelWakeupThread( ) { MipsCpu.halt = 1; } 
void SYSCALL_SuspendThread( ) { MipsCpu.halt = 1; } 
void SYSCALL_ResumeThread( ) { MipsCpu.halt = 1; } 
void SYSCALL_JoinThread( ) { MipsCpu.halt = 1; } 

void	SYSCALL_RFU060() {
			
		MipsCpu.GPR[CPU_REG_V0].lo = MipsCpu.GPR[CPU_REG_SP].lo ;				
		MipsCpu.PC = MipsCpu.PC + 4;

}

void	SYSCALL_RFU061() {
				
		MipsCpu.GPR[CPU_REG_V0].lo =	MipsCpu.GPR[CPU_REG_A0].lo + 	MipsCpu.GPR[CPU_REG_A1].lo ;						//PCSX2 hack
		MipsCpu.PC = MipsCpu.PC + 4;

}

void SYSCALL_EndOfHeap( ) { MipsCpu.halt = 1; } 
void SYSCALL_RFU063( ) { MipsCpu.halt = 1; } 

void SYSCALL_CreateSema( ) { 

	MipsCpu.GPR[2].lo = 0x0 ;						//PCSX2 hack

	MipsCpu.PC = MipsCpu.PC + 4;	
	//MipsCpu.halt = 1; 
} 

void SYSCALL_DeleteSema( ) {
 
 MipsCpu.PC = MipsCpu.PC + 4;	
 
 } 
void SYSCALL_SignalSema( ) { MipsCpu.halt = 1; } 
void SYSCALL_WaitSema( ) { MipsCpu.halt = 1; } 
void SYSCALL_PollSema( ) { MipsCpu.halt = 1; } 
void SYSCALL_ReferSemaStatus( ) { MipsCpu.halt = 1; } 
void SYSCALL_RFU073( ) { MipsCpu.halt = 1; } 
void SYSCALL_SetOsdConfigParam( ) { MipsCpu.halt = 1; } 
void SYSCALL_GetOsdConfigParam( ) { MipsCpu.halt = 1; } 
void SYSCALL_GetGsHParam( ) { MipsCpu.halt = 1; } 
void SYSCALL_GetGsVParam( ) { MipsCpu.halt = 1; } 
void SYSCALL_SetGsHParam( ) { MipsCpu.halt = 1; } 
void SYSCALL_SetGsVParam( ) { MipsCpu.halt = 1; } 
void SYSCALL_RFU080_CreateEventFlag( ) { MipsCpu.halt = 1; } 
void SYSCALL_RFU081_DeleteEventFlag( ) { MipsCpu.halt = 1; } 
void SYSCALL_RFU082_SetEventFlag( ) { MipsCpu.halt = 1; } 
void SYSCALL_RFU083_iSetEventFlag( ) { MipsCpu.halt = 1; } 
void SYSCALL_RFU084_ClearEventFlag( ) { MipsCpu.halt = 1; } 
void SYSCALL_RFU085_iClearEventFlag( ) { MipsCpu.halt = 1; } 
void SYSCALL_RFU086_WaitEvnetFlag( ) { MipsCpu.halt = 1; } 
void SYSCALL_RFU087_PollEvnetFlag( ) { MipsCpu.halt = 1; } 
void SYSCALL_RFU088_iPollEvnetFlag( ) { MipsCpu.halt = 1; } 
void SYSCALL_RFU089_ReferEventFlagStatus( ) { MipsCpu.halt = 1; } 
void SYSCALL_RFU091( ) { MipsCpu.halt = 1; } 
void SYSCALL_EnableIntcHandler( ) { MipsCpu.halt = 1; } 
void SYSCALL_DisableIntcHandler( ) { MipsCpu.halt = 1; } 
void SYSCALL_EnableDmacHandler( ) { MipsCpu.halt = 1; } 
void SYSCALL_DisableDmacHandler( ) { MipsCpu.halt = 1; } 
void SYSCALL_KSeg0( ) { MipsCpu.halt = 1; } 
void SYSCALL_EnableCache( ) { MipsCpu.halt = 1; } 
void SYSCALL_DisableCache( ) { MipsCpu.halt = 1; } 
void SYSCALL_GetCop0( ) { MipsCpu.halt = 1; } 

void SYSCALL_FlushCache( ) {
	
	MipsCpu.PC = MipsCpu.PC + 4;
 } 

void SYSCALL_CpuConfig( ) { MipsCpu.halt = 1; } 
void SYSCALL_sceSifStopDma( ) { MipsCpu.halt = 1; } 
void SYSCALL_SetCPUTimerHandler( ) { MipsCpu.halt = 1; } 
void SYSCALL_SetCPUTimer( ) { MipsCpu.halt = 1; } 
void SYSCALL_ForceRead( ) { MipsCpu.halt = 1; } 
void SYSCALL_ForceWrite( ) { MipsCpu.halt = 1; } 

void SYSCALL_GsGetIMR( ) { 


	MipsCpu.PC = MipsCpu.PC + 4;
} 

void SYSCALL_GsPutIMR( ) { 
	
	
	MipsMemoryWrite ( MipsCpu.GPR[CPU_REG_A0] , DATA_DW ,  0x12001010  ,0 ) ;			// (12001010 : GS_IMR) = A0
	
	MipsCpu.PC = MipsCpu.PC + 4;

} 

void SYSCALL_SetPgifHandler( ) { MipsCpu.halt = 1; } 
void SYSCALL_SetVSyncFlag( ) { MipsCpu.halt = 1; } 
void SYSCALL_RFU116( ) { MipsCpu.halt = 1; } 
void SYSCALL_print( ) { MipsCpu.halt = 1; } 
void SYSCALL_sceSifDmaStat( ) { MipsCpu.halt = 1; } 
void SYSCALL_sceSifSetDma( ) { MipsCpu.halt = 1; } 

void SYSCALL_sceSifSetDChain( )  {
//	MipsCpu.halt = 1;
	MipsCpu.PC = MipsCpu.PC + 4;
 } 
 


void SYSCALL_sceSifSetReg( ) { MipsCpu.halt = 1; } 

void SYSCALL_sceSifGetReg( ) {
	
	MipsCpu.GPR[2].lo = 0x00060000;
	

	MipsCpu.PC = MipsCpu.PC + 4;
 } 
 
void SYSCALL_ExecOSD( ) { MipsCpu.halt = 1; } 
void SYSCALL_Deci2Call( ) { MipsCpu.halt = 1; } 
void SYSCALL_PSMode( ) { MipsCpu.halt = 1; } 
void SYSCALL_MachineType( ) { MipsCpu.halt = 1; } 
void SYSCALL_GetMemorySize( ) { MipsCpu.halt = 1; } 
void SYSCALL_isceSifSetDChain( ) { MipsCpu.halt = 1; } 
void SYSCALL_isceSifSetDma( ) { MipsCpu.halt = 1; } 
void SYSCALL_isceSifDmaStat( ) { MipsCpu.halt = 1; } 

void SYSCALL_iGsPutIMR( ) { 
	MipsCpu.PC = MipsCpu.PC + 4;
} 

void SYSCALL_iGsGetIMR( ) {
	MipsCpu.PC = MipsCpu.PC + 4;
 } 
void SYSCALL_iCpuConfig( ) { MipsCpu.halt = 1; } 
void SYSCALL_iFlushCache( ) { MipsCpu.halt = 1; } 
void SYSCALL_iGetCop0( ) { MipsCpu.halt = 1; } 
void SYSCALL_iDisableDmacHandler( ) { MipsCpu.halt = 1; } 
void SYSCALL_iEnableDmacHandler( ) { MipsCpu.halt = 1; } 
void SYSCALL_iDisableIntcHandler( ) { MipsCpu.halt = 1; } 
void SYSCALL_iEnableIntcHandler( ) { MipsCpu.halt = 1; } 
void SYSCALL_RFU090_iReferEventFlagStatus( ) { MipsCpu.halt = 1; } 
void SYSCALL_iReferSemaStatus( ) { MipsCpu.halt = 1; } 
void SYSCALL_iPollSema( ) { MipsCpu.halt = 1; } 
void SYSCALL_iSignalSema( ) { MipsCpu.halt = 1; } 
void SYSCALL_iResumeThread( ) { MipsCpu.halt = 1; } 
void SYSCALL_iSuspendThread( ) { MipsCpu.halt = 1; } 
void SYSCALL_iCancelWakeupThread( ) { MipsCpu.halt = 1; } 
void SYSCALL_iWakeupThread( ) { MipsCpu.halt = 1; } 
void SYSCALL_iReferThreadStatus( ) { MipsCpu.halt = 1; } 
void SYSCALL_iReleaseWaitThread( ) { MipsCpu.halt = 1; } 
void SYSCALL_iRotateThreadReadyQueue( ) { MipsCpu.halt = 1; } 
void SYSCALL_iChangeThreadPriority( ) { MipsCpu.halt = 1; } 
void SYSCALL_iTerminateThread( ) { MipsCpu.halt = 1; } 
void SYSCALL__iReleaseAlarm( ) { MipsCpu.halt = 1; } 
void SYSCALL__iSetAlarm( ) { MipsCpu.halt = 1; } 
void SYSCALL__iDisableDmac( ) { MipsCpu.halt = 1; } 
void SYSCALL__iEnableDmac( ) { MipsCpu.halt = 1; } 
void SYSCALL__iDisableIntc( ) { MipsCpu.halt = 1; } 
void SYSCALL__iEnableIntc( ) { MipsCpu.halt = 1; } 

void	MADD( ){ MipsCpu.halt = 1; } 

		
void	MADDU( ){ MipsCpu.halt = 1; } 
void	MUL( ){ MipsCpu.halt = 1; } 
void	MSUB( ){ MipsCpu.halt = 1; } 
void	MSUBU( ){ MipsCpu.halt = 1; } 
void	CLZ( ){ MipsCpu.halt = 1; } 
void	CLO( ){ MipsCpu.halt = 1; } 
void	DLCZ( ) { MipsCpu.halt = 1; } 
void	DLCO( ){ MipsCpu.halt = 1; } 
void	SDBBP( ){ MipsCpu.halt = 1; } 

void	MULT1( )	{
	_u64 hi_temp,lo_temp;
	

	Extract_rs();
	Extract_rt();
	Extract_rd();			

	if ( debug_op == 1 )printf("	mult	%s, %s, %s \n", debug_r5900_reg[rd] , debug_r5900_reg[rs],debug_r5900_reg[rt] );

	lo_temp = (( MipsCpu.GPR[rs].lo & 0xFFFFFFFF ) * ( MipsCpu.GPR[rt].lo & 0xFFFFFFFF )) ;
	hi_temp = ( (MipsCpu.GPR[rs].lo >> 32 ) & 0xFFFFFFFF ) * ( (MipsCpu.GPR[rt].lo >> 32 )& 0xFFFFFFFF );
	
	MipsCpu.GPR[CPU_REG_LO].hi = lo_temp;
	MipsCpu.GPR[CPU_REG_HI].hi = hi_temp;
	
	MipsCpu.GPR[rd].lo = lo_temp ;
	
	
	MipsCpu.PC = MipsCpu.PC + 4;
	
}


void	MFC1(){ MipsCpu.halt = 1; }
void	DMFC1(){ MipsCpu.halt = 1; }	
void	CFC1(){ MipsCpu.halt = 1; }			
void	MTC1(){ MipsCpu.halt = 1; }	
void	DMTC1(){ MipsCpu.halt = 1; }	
void	CTC1(){ MipsCpu.halt = 1; }	
void	BC1(){ MipsCpu.halt = 1; }	 		
void	COP1_S(){ MipsCpu.halt = 1; };
void	COP1_D(){ MipsCpu.halt = 1; }	

void	COP1_W(){ 	cop1_W_L_opc[MipsCpu.opcode&0x3F]();	}

void	COP1_L(){ MipsCpu.halt = 1; }




void	BC1F(){ MipsCpu.halt = 1; } 
void	BC1T(){ MipsCpu.halt = 1; }
void	BC1FL(){ MipsCpu.halt = 1; }
void	BC1TL(){ MipsCpu.halt = 1; }

void	ADD_S(){ MipsCpu.halt = 1; } 		
void	SUB_S(){ MipsCpu.halt = 1; } 		
void	MUL_S(){ MipsCpu.halt = 1; } 		
void	DIV_S(){ MipsCpu.halt = 1; } 		
void	SQRT_S(){ MipsCpu.halt = 1; } 	
void	ABS_S(){ MipsCpu.halt = 1; } 		
void	MOV_S(){ MipsCpu.halt = 1; } 		
void	NEG_S(){ MipsCpu.halt = 1; }
void	ROUND_L_S(){ MipsCpu.halt = 1; }	
void	TRUNC_L_S(){ MipsCpu.halt = 1; }	
void	CEIL_L_S(){ MipsCpu.halt = 1; }	
void	FLOOR_L_S(){ MipsCpu.halt = 1; }	
void	ROUND_W_S(){ MipsCpu.halt = 1; }	
void	TRUNC_W_S(){ MipsCpu.halt = 1; }	
void	CEIL_W_S(){ MipsCpu.halt = 1; }	
void	FLOOR_W_S(){ MipsCpu.halt = 1; }
void	MOVCF_S(){ MipsCpu.halt = 1; }	
void	MOVZ_S(){ MipsCpu.halt = 1; }		
void	MOVN_S(){ MipsCpu.halt = 1; }		
void	RECIP_S(){ MipsCpu.halt = 1; }  	
void	RSQRT_S(){ MipsCpu.halt = 1; }  	
void	CVT_D_S(){ MipsCpu.halt = 1; }		
void	CVT_W_S(){ MipsCpu.halt = 1; }	
void	CVT_L_S(){ MipsCpu.halt = 1; }	
void	C_F_S(){ MipsCpu.halt = 1; } 		
void	C_UN_S(){ MipsCpu.halt = 1; }		
void	C_EQ_S(){ MipsCpu.halt = 1; }		
void	C_UEQ_S(){ MipsCpu.halt = 1; }	
void	C_OLT_S(){ MipsCpu.halt = 1; }	
void	C_ULT_S(){ MipsCpu.halt = 1; }	
void	C_OLE_S(){ MipsCpu.halt = 1; }	
void	C_ULE_S(){ MipsCpu.halt = 1; }
void	C_SF_S(){ MipsCpu.halt = 1; }		
void	C_NGLE_S(){ MipsCpu.halt = 1; } 	
void	C_SEQ_S(){ MipsCpu.halt = 1; }	
void	C_NGL_S(){ MipsCpu.halt = 1; }	
void	C_LT_S(){ MipsCpu.halt = 1; }		
void	C_NGE_S(){ MipsCpu.halt = 1; }	
void	C_LE_S(){ MipsCpu.halt = 1; }		
void	C_NGT_S(){ MipsCpu.halt = 1; }



void	ADD_D(){ MipsCpu.halt = 1; }		
void	SUB_D(){ MipsCpu.halt = 1; }		
void	MUL_D(){ MipsCpu.halt = 1; }		
void	DIV_D(){ MipsCpu.halt = 1; }		
void	SQRT_D(){ MipsCpu.halt = 1; }		
void	ABS_D(){ MipsCpu.halt = 1; }		
void	MOV_D(){ MipsCpu.halt = 1; }		
void	NEG_D(){ MipsCpu.halt = 1; }
void	ROUND_L_D(){ MipsCpu.halt = 1; }	
void	TRUNC_L_D(){ MipsCpu.halt = 1; }	
void	CEIL_L_D(){ MipsCpu.halt = 1; }	
void	FLOOR_L_D(){ MipsCpu.halt = 1; }	
void	ROUND_W_D(){ MipsCpu.halt = 1; }	
void	TRUNC_W_D(){ MipsCpu.halt = 1; }	
void	CEIL_W_D(){ MipsCpu.halt = 1; }	
void	FLOOR_W_D(){ MipsCpu.halt = 1; }
void	MOVCF_D(){ MipsCpu.halt = 1; }	
void	MOVZ_D(){ MipsCpu.halt = 1; }		
void	MOVN_D(){ MipsCpu.halt = 1; }				
void	RECIP_D(){ MipsCpu.halt = 1; }	
void	RSQRT_D(){ MipsCpu.halt = 1; }
void	CVT_S_D(){ MipsCpu.halt = 1; }	
void	CVT_W_D(){ MipsCpu.halt = 1; }	
void	CVT_L_D(){ MipsCpu.halt = 1; }	
void	C_F_D(){ MipsCpu.halt = 1; }		
void	C_UN_D(){ MipsCpu.halt = 1; }		
void	C_EQ_D(){ MipsCpu.halt = 1; }		
void	C_UEQ_D(){ MipsCpu.halt = 1; }	
void	C_OLT_D(){ MipsCpu.halt = 1; }	
void	C_ULT_D(){ MipsCpu.halt = 1; }	
void	C_OLE_D(){ MipsCpu.halt = 1; }	
void	C_ULE_D(){ MipsCpu.halt = 1; }
void	C_SF_D(){ MipsCpu.halt = 1; }		
void	C_NGLE_D(){ MipsCpu.halt = 1; }	
void	C_SEQ_D(){ MipsCpu.halt = 1; }	
void	C_NGL_D(){ MipsCpu.halt = 1; }	
void	C_LT_D(){ MipsCpu.halt = 1; }		
void	C_NGE_D(){ MipsCpu.halt = 1; }	
void	C_LE_D(){ MipsCpu.halt = 1; }		
void	C_NGT_D(){ MipsCpu.halt = 1; }


void	CVT_S(){ 
	
	MipsCpu.halt = 1; 
	}		
void	CVT_D(){ MipsCpu.halt = 1; }			


void	MOVF_fmt(){ MipsCpu.halt = 1; } 
void	MOVT_fmt(){ MipsCpu.halt = 1; }	

