/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: muaddib_atreides@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdio.h>
#include <stdlib.h>

#include "nSX2.h"
#include "elf_file.h"

#include "plugin_gs/PS2Edefs.h"

_u8 	* MipsMemory;
_u8	* SpecialAreaMemory;
_u8	* GSAreaMemory;
_u8   * BiosMemory;


/****************************************************************************************************/

_u8 MipsMemoryAlloc( void )  {
	_u8 	Result;	
	printf("R5900:Allocation Main Memory 32 Mbits...");
	MipsMemory = malloc ( sizeof( _u8 ) * 0x01FFFFFF );
	if (MipsMemory != NULL ) {
		printf(" OK \n");
		Result = TRUE;
	}
	else {
		printf(" KO \n");
		Result = FALSE;
	}
	return Result;
}

_u8 MipsSpecialAreaMemoryAlloc( void )  {
	_u8 	Result;	
	printf("R5900:Allocation Special Memory 64 kbytes...");
	SpecialAreaMemory = malloc ( sizeof( _u8 ) * 0xFFFF );
	if (SpecialAreaMemory != NULL ) {
		printf(" OK \n");
		Result = TRUE;
	}
	else {
		printf(" KO \n");
		Result = FALSE;
	}
	return Result;
}

_u8 MipsGSAreaMemoryAlloc( void )  {
	_u8 	Result;	
	printf("R5900:Allocation GS Register Memory 4288 bytes...");
	GSAreaMemory = malloc ( sizeof( _u8 ) * 0x10C0 );
	if (GSAreaMemory != NULL ) {
		printf(" OK \n");
		Result = TRUE;
	}
	else {
		printf(" KO \n");
		Result = FALSE;
	}
	return Result;
}

_u8 MipsBiosMemoryAlloc( void )  {
	_u8 	Result;	
	printf("R5900:Allocation BIOS Memory 4 M bytes...");
	BiosMemory = malloc ( sizeof( _u8 ) * 0x3FFFFF );
	if (BiosMemory != NULL ) {
		printf(" OK \n");
		Result = TRUE;
	}
	else {
		printf(" KO \n");
		Result = FALSE;
	}
	return Result;
}



void IcoFreeMemory( void ) {
	printf("FreeMemory: Main Memory\n");
	free( MipsMemory );
	free(SpecialAreaMemory);
	free(BiosMemory);
}

/****************************************************************************************************/


void CopyBiosToMemory (   ) {

		_u64 Addr;
		char c;
		
		Addr = 0;
		printf("Init:Copy BIOS To Memory ...");
		fflush( stdout );
		
		
		fseek( BiosFile, 0, SEEK_SET );
		while ( fread( &c , sizeof( char ), 1, BiosFile ) != 0 ) {  
			BiosMemory[Addr] = c;
			Addr++;
		}
		printf("Done\n");
}


void CopyProgramToMemory( void ) {
	
	_u64 Addr;
	_u64 fIndex;
	_u64	Index2;
	char c;
	
	printf("Init:Copy Program To Memory:");
	
	fflush( stdout );
	
	for ( Index2 = 0 ; Index2 <  ico_elf_hdr.num_of_prg_hdr ; Index2++ ) {
			printf(".");
			fflush( stdout );
	
		Addr = ico_elf_prg_hdr[Index2].physic_add_of_segment;
		fseek( ElfFile,  ico_elf_prg_hdr[Index2].offs_from_start_to_prg_seg , SEEK_SET );	
	
		for ( fIndex = 0 ; fIndex < ico_elf_prg_hdr[Index2].num_of_bytes_in_file_ima ; fIndex++ ) {
	
			if ( fread( &c , sizeof( char ), 1, ElfFile ) != 0 ) {
				MipsMemory[Addr] = c;
				Addr++;
			}
		
		}
	}
	printf("Done\n");

}

void CopySectionToMemory ( void ) {

	_u64 Addr;
	_u8 index;
	_u64 size;
	char c;
	
	printf("Init:Copy Section To Memory:");
	fflush( stdout );
		
	for ( index = 0 ; index <  ico_elf_hdr.num_of_sec_hdr ; index++ ) {
			
			
			if ( ( ico_elf_sec_hdr[index].sect_start_add +  ico_elf_sec_hdr[index].size ) < 0x01FFFFFF ) {
				printf(".");
				fflush( stdout );
			
				Addr =  ico_elf_sec_hdr[index].sect_start_add,
				fseek( ElfFile,   ico_elf_sec_hdr[index].offset_from_start  , SEEK_SET );	
			
				for ( size = 0 ; size < ico_elf_sec_hdr[index].size ; size++ ) {
				
					if ( fread( &c , sizeof( char ), 1, ElfFile ) != 0 ) {
						MipsMemory[Addr] = c;
						Addr++;
					}
			
				}
			
			}
			else {
				printf(" ");
				fflush( stdout );
			}
			
	
	}

	printf("Done\n");

}

/***********************************************************************************************/

void	MipsMemoryWrite ( _u128 fVal  , _u8 fDataType , _u64 fBase , int fOffset ) {
	
	_u8	Mem_Perform;
	_u64	PhysicalAdd;
	
	_u8 * MemTarget;
	
	Mem_Perform = 0;
	PhysicalAdd = ( fBase + fOffset ) & 0xFFFFFFFF;
	
	//	0x10000000 - 0x1000FFFF	
	if ( ( PhysicalAdd >= 0x10000000 ) && ( PhysicalAdd <= 0x1000FFFF ) ){
		MemTarget = SpecialAreaMemory;
		MipsSpecialMemWrite( MemTarget , fVal ,  fDataType , fBase , fOffset );
		Mem_Perform = 1;
		return;
	}


	//	0x12000000 - 0x12001080	
	if ( ( PhysicalAdd >= 0x12000000 ) && ( PhysicalAdd <= 0x12001080 ) ){
		MemTarget = GSAreaMemory;
		MipsGSAreaMemWrite( MemTarget, fVal ,  fDataType , fBase , fOffset );
		Mem_Perform = 1;
		return;
	}

	
	
	// 0x80000000 - 0x9FFFFFFF			//KSEG0
	if  (( PhysicalAdd >= 0x80000000 ) && ( PhysicalAdd < 0x9FFFFFFF ) ) {
		//	if (PhysicalAdd >=  0x81effc70 ) printf("Mem:WRITE %16.16llx PC:%8.8lx\n",PhysicalAdd, MipsCpu.PC);
		PhysicalAdd = PhysicalAdd & 0x01FFFFFF;
	}
	
	//	0x00100000 - 0x01FFFFFF
	//	0x20100000 - 0x21FFFFFF
	//	0x30100000 - 0x31FFFFFF
	
		if ( (( PhysicalAdd >= 0x00100000 ) && ( PhysicalAdd < 0x01FFFFFF )) ||
	 		(( PhysicalAdd >= 0x20100000 ) && ( PhysicalAdd < 0x21FFFFFF )) ||
	 		(( PhysicalAdd >= 0x30100000 ) && ( PhysicalAdd < 0x31FFFFFF )) )
	 {
		
		PhysicalAdd = PhysicalAdd &	0x01FFFFFF;

		MemTarget = MipsMemory;
		
		switch ( fDataType ) {
			case DATA_QUAD : {
												MipsMemoryWriteDW ( MemTarget , fVal.lo , PhysicalAdd );
												MipsMemoryWriteDW ( MemTarget, fVal.hi , PhysicalAdd + 8 );
												break;
										}
			case DATA_DW : {
												MipsMemoryWriteDW (MemTarget , fVal.lo , PhysicalAdd );
												break;
										}

			case DATA_DWH : {
												MipsMemoryWriteDW (MemTarget , fVal.hi , PhysicalAdd );
												break;
										}
			
			
			case DATA_W : {
												MipsMemoryWriteW ( MemTarget , fVal.lo , PhysicalAdd );
												break;
										}	
			case DATA_HW : {
												MipsMemoryWriteHW ( MemTarget , fVal.lo , PhysicalAdd );
												break;
										}				
			
			
			case DATA_B : {
												MipsMemoryWriteB ( MemTarget ,fVal.lo , PhysicalAdd );
												break;
										}																	
		
			default: 	{
								printf("Mem:Write DataType Not Handled\n");
								MipsCpu.halt = 1;
								break;
							}
		}
		Mem_Perform = 1;
		return;
	}

	// Others
	if ( Mem_Perform == 0 ) {
		printf("Mem:Write:NEED SPECIAL HANDLE ADD :  %llx\n",PhysicalAdd );
		MipsCpu.halt = 1;
	}


}


void	MipsMemoryWriteDW ( _u8 * Mem ,  _u64 fVal , _u64 fAdd ) {
	

	
	Mem[ fAdd + 0] = ( fVal & 0x00000000000000FF ) ;
	Mem[ fAdd + 1] = ( fVal & 0x000000000000FF00 ) >> 8 ;
	Mem[ fAdd + 2] = ( fVal & 0x0000000000FF0000 ) >> 16 ;
	Mem[ fAdd + 3] = ( fVal & 0x00000000FF000000 ) >> 24 ;
	Mem[ fAdd + 4] = ( fVal & 0x000000FF00000000 ) >> 32 ;
	Mem[ fAdd + 5] = ( fVal & 0x0000FF0000000000 ) >> 40 ;
	Mem[ fAdd + 6] = ( fVal & 0x00FF000000000000 ) >> 48 ;
	Mem[ fAdd + 7] = ( fVal & 0xFF00000000000000 ) >> 56 ;
}

void MipsMemoryWriteW  (_u8 * Mem ,  _u64 fVal , _u64 fAdd ) {

	
	Mem[ fAdd + 0] = ( fVal & 0x00000000000000FF ) ;
	Mem[ fAdd + 1] = ( fVal & 0x000000000000FF00 ) >> 8 ;
	Mem[ fAdd + 2] = ( fVal & 0x0000000000FF0000 ) >> 16 ;
	Mem[ fAdd + 3] = ( fVal & 0x00000000FF000000 ) >> 24 ;
								
}

void MipsMemoryWriteHW  ( _u8 * Mem , _u64 fVal , _u64 fAdd ) {

	
	Mem[ fAdd + 0] = ( fVal & 0x00000000000000FF ) ;
	Mem[ fAdd + 1] = ( fVal & 0x000000000000FF00 ) >> 8 ;
								
}

void MipsMemoryWriteB  ( _u8 * Mem ,_u64 fVal , _u64 fAdd ) {

	
	Mem[ fAdd + 0] = ( fVal & 0x00000000000000FF ) ;
								
}

/******************************************************************************************/

void	MipsMemoryRead ( _u128 * fVal  , _u8 fDataType , _u64 fBase , int fOffset ) {

	_u8	Mem_Perform;
	_u64	PhysicalAdd;
	
	_u8 * MemTarget;
	
	Mem_Perform = 0;
	PhysicalAdd = ( fBase + fOffset ) & 0xFFFFFFFF;
	
	//printf("Mem:Read %16.16llx %8.8lx\n",PhysicalAdd, MipsCpu.PC);
		
	//	0x10000000 - 0x1000FFFF	
	if ( ( PhysicalAdd >= 0x10000000 ) && ( PhysicalAdd < 0x1000FFFF ) ){
		MemTarget = SpecialAreaMemory;
		MipsSpecialMemRead(MemTarget , fVal ,  fDataType , fBase , fOffset );
		Mem_Perform = 1;
		return;
	}

		//	0x12000000 - 0x12001080	
	if ( ( PhysicalAdd >= 0x12000000 ) && ( PhysicalAdd <= 0x12001080 ) ){
		MemTarget = GSAreaMemory;
		MipsGSAreaMemRead( MemTarget,  fVal ,  fDataType , fBase , fOffset );
		Mem_Perform = 1;
		return;
	}

	// 0x1FC00000 - 0x1FFFFFFF	ROM BIOS Mirror Adress.	(not cached)
	// 0x9FC00000 - 0x9FFFFFFF	ROM BIOS Mirror Adress.	(cached)
	//0xBFC00000 - 0xBFFFFFFF	ROM BIOS Mirror Adress.	(not cached)
	if ( 	(( PhysicalAdd >= 0x1FC00000 ) && ( PhysicalAdd < 0x1FFFFFFF )) ||
	 		(( PhysicalAdd >= 0x9FC00000 ) && ( PhysicalAdd < 0x9FFFFFFF )) ||
	 		(( PhysicalAdd >= 0xBFC00000 ) && ( PhysicalAdd < 0xBFFFFFFF )) )
	 {

		MemTarget = BiosMemory; 
	 	MipsBiosAreaMemRead(  MemTarget , fVal ,  fDataType , fBase , fOffset );

		Mem_Perform = 1;
		return;
	 
	 }


	// 0x80000000 - 0x9FFFFFFF			//KSEG0
	if  (( PhysicalAdd >= 0x80000000 ) && ( PhysicalAdd < 0x9FFFFFFF ) ) {
	//	if (PhysicalAdd >=  0x81effc70 ) printf("Mem:READ  %16.16llx PC:%8.8lx\n",PhysicalAdd, MipsCpu.PC);
		PhysicalAdd = PhysicalAdd & 0x01FFFFFF;

	}


	//	0x00100000 - 0x01FFFFFF
	//	0x20100000 - 0x21FFFFFF
	//	0x30100000 - 0x31FFFFFF
	if ( 	(( PhysicalAdd >= 0x00100000 ) && ( PhysicalAdd < 0x01FFFFFF )) ||
	 		(( PhysicalAdd >= 0x20100000 ) && ( PhysicalAdd < 0x21FFFFFF )) ||
	 		(( PhysicalAdd >= 0x30100000 ) && ( PhysicalAdd < 0x31FFFFFF )) )
	 {
		
		PhysicalAdd = PhysicalAdd &	0x01FFFFFF;
		MemTarget = MipsMemory; 
		switch ( fDataType ) {
		
			
			case DATA_DWH : {
												MipsMemoryReadDWH (MemTarget , fVal , PhysicalAdd );
												break;
										}
			
			case DATA_DW : {
												MipsMemoryReadDW ( MemTarget, fVal , PhysicalAdd );
												break;
										}
		
				
			case DATA_W : {
												MipsMemoryReadW ( MemTarget, fVal , PhysicalAdd );
												break;
										}
			
			case DATA_HW : {
												MipsMemoryReadHW ( MemTarget, fVal , PhysicalAdd );
												break;
										}

			case DATA_B : {
												MipsMemoryReadB ( MemTarget , fVal , PhysicalAdd );
												break;
										}										
			default: 	{
								printf("Mem:Read DataType Not Handled\n");
								MipsCpu.halt = 1;
								break;
							}
		}
		Mem_Perform = 1;
		return;
	}

	// Others
	if ( Mem_Perform == 0 ) {
		printf("Mem:Read:NEED SPECIAL HANDLE ADD :  %llx\n",PhysicalAdd );
		MipsCpu.halt = 1;
	}

}




void MipsMemoryReadDW ( _u8 * Mem,  _u128 * fVal , _u64 fAddr ) {

	_u64 temp  ;
	temp	 = 	 	( Mem[ fAddr + 0]  & 0xFF  )   				|
						(  ( Mem[ fAddr + 1]  << 8 ) 	& 0xFF00   )  |
						( ( Mem[ fAddr + 2]   << 16 )  	& 0xFF0000 ) | 
						( ( Mem[ fAddr + 3]   << 24 )  	& 0xFF000000 ) |
						( ( Mem[ fAddr + 4]   << 32 )  	& 0xFF00000000 ) | 
						(  ( Mem[ fAddr + 5] << 40 )  	& 0xFF0000000000 ) | 
						( ( Mem[ fAddr + 6]  << 48 )  	& 0xFF000000000000 ) | 
						( (  Mem[ fAddr + 7] << 56 ) 	& 0xFF00000000000000 );

	fVal->lo = temp;
}

void MipsMemoryReadDWH ( _u8 * Mem,  _u128 * fVal , _u64 fAddr ) {

	_u64 temp  ;
	temp	 = 	 	( Mem[ fAddr + 0]  & 0xFF  )   				|
						(  ( Mem[ fAddr + 1]  << 8 ) 	& 0xFF00   )  |
						( ( Mem[ fAddr + 2]   << 16 )  	& 0xFF0000 ) | 
						( ( Mem[ fAddr + 3]   << 24 )  	& 0xFF000000 ) |
						( ( Mem[ fAddr + 4]   << 32 )  	& 0xFF00000000 ) | 
						(  ( Mem[ fAddr + 5] << 40 )  	& 0xFF0000000000 ) | 
						( ( Mem[ fAddr + 6]  << 48 )  	& 0xFF000000000000 ) | 
						( (  Mem[ fAddr + 7] << 56 ) 	& 0xFF00000000000000 );


	fVal->hi = temp;
}


void MipsMemoryReadW ( _u8 * Mem, _u128 * fVal , _u64 fAddr ) {

	_u64 temp  ;
	temp	 = 	 	( Mem[ fAddr + 0]  & 0xFF  )   				|
						(  ( Mem[ fAddr + 1]  << 8 ) 	& 0xFF00   )  |
						( ( Mem[ fAddr + 2]   << 16 )  	& 0xFF0000 ) | 
						( ( Mem[ fAddr + 3]   << 24 )  	& 0xFF000000 ) ;
	

	fVal->lo = temp;
}

void MipsMemoryReadHW ( _u8 * Mem,  _u128 * fVal , _u64 fAddr ) {

	_u64 temp  ;
	temp	 = 	 	( Mem[ fAddr + 0]  & 0xFF  )   				|
						(  ( Mem[ fAddr + 1]  << 8 ) 	& 0xFF00   )  ;
	

	fVal->lo = temp;
	

}


void MipsMemoryReadB ( _u8 * Mem,  _u128 * fVal , _u64 fAddr ) {

	_u64 temp  ;
	temp	 = 	 	( Mem[ fAddr + 0]  & 0xFF  )   ;					
	fVal->lo = temp;
	

}


/*****************************************************************************************************/

void  MipsSpecialMemRead( _u8 * Mem, _u128 * fVal  , _u8 fDataType , _u64 fBase , int fOffset ) {

	_u64 RealAdd;
	RealAdd = ( fBase + fOffset ) & 0xFFFF;





	switch ( fDataType ) {

						
			case DATA_DW : {
												MipsMemoryReadDW ( Mem, fVal , RealAdd );
												break;
										}
				
			case DATA_W : {
												MipsMemoryReadW ( Mem, fVal , RealAdd );
												break;
										}

																	
			default: 	{
								printf("MemSpecial:Read DataType Not Handled\n");
								MipsCpu.halt = 1;
								break;
							}
		}


	if ( RealAdd >= 0x8000 ) {
		Ps2Bios.ReadToDmaArea = 1;
		Ps2Bios.DmaAreaReadAt = RealAdd;
		Ps2Bios.DmaAreaReadVal = fVal->lo;
	}

}


void  MipsSpecialMemWrite( _u8 * Mem , _u128  fVal  , _u8 fDataType , _u64 fBase , int fOffset ) {

	_u64 RealAdd;
	RealAdd = ( fBase + fOffset ) & 0xFFFF;

	/* Can't write to DMA Status */
	if ( RealAdd == 0xE010 ) 		return;			
	
	if ( RealAdd >= 0x8000 ) {
		Ps2Bios.WriteToDmaArea = 1;
		Ps2Bios.DmaAreaWriteAt = RealAdd;
		Ps2Bios.DmaAreaWriteVal = fVal.lo;
	}
	

	switch ( fDataType ) {
				
			case DATA_DW : {
												MipsMemoryWriteDW ( Mem , fVal.lo , RealAdd );
												break;
										}	
				
			case DATA_W : {
												MipsMemoryWriteW ( Mem , fVal.lo , RealAdd );
												break;
										}
																	
			default: 	{
								printf("MemSpecial:Write DataType Not Handled\n");
								MipsCpu.halt = 1;
								break;
							}
		}

}



/***********************************************************************************************/



void  MipsGSAreaMemWrite( _u8 * Mem , _u128  fVal  , _u8 fDataType , _u64 fBase , int fOffset ) {

	_u64 RealAdd;
	_u32 Add;
//	RealAdd = ( fBase + fOffset ) & 0xFFFF;
	Add = ( fBase + fOffset ) ;
	
//	Ps2Bios.WriteToGSArea = 1;
//	Ps2Bios.GSAreaWriteAt = RealAdd;
//	Ps2Bios.GSAreaWriteVal = fVal.lo;
	

	switch ( fDataType ) {
			
			case DATA_DW : {	
								if ( debug_gs == 1 ) printf("%8.8llx: GSwrite64 at %lx with %llx\n",MipsCpu.PC,Add,fVal.lo);
								GSwrite64(Add, fVal.lo) ;				//and pass the value to the GS pluggin
								break;
							}

			case DATA_W : {
								if ( debug_gs == 1 ) printf("%8.8llx: GSwrite32 at %lx with %llx\n",MipsCpu.PC,Add,fVal.lo);
								GSwrite32(Add,(_u32) (fVal.lo&0x00000000FFFFFFFF)) ;		//and pass the value to the GS pluggin
								break;
						}

																	
			default: 	{
								printf("MemGSArea:Write DataType Not Handled\n");
								MipsCpu.halt = 1;
								break;
							}
		}
}



void MipsGSAreaMemRead( _u8 * Mem, _u128 * fVal  , _u8 fDataType , _u64 fBase , int fOffset ) {

	_u64 RealAdd;
	RealAdd = ( fBase + fOffset ) ; //& 0xFFFF;

	switch ( fDataType ) {

				
			case DATA_DW : {
		
								fVal->lo = GSread64((_u32)RealAdd) ;
								if ( debug_gs == 1 ) printf("%8.8llx: GSread64 at %llx with %llx\n",MipsCpu.PC,RealAdd,fVal->lo);
								
								break;
							}
	
			case DATA_W : {

								fVal->lo = GSread32((_u32)RealAdd) ;
								if ( debug_gs == 1 ) printf("%8.8llx: GSread32 at %llx with %llx\n",MipsCpu.PC,RealAdd,fVal->lo);

								break;
							}
			
																	
			default: 	{
								printf("MemGSArea:Read DataType Not Handled\n");
								MipsCpu.halt = 1;
								break;
							}
		}

}

/*****************************************************************************************************/


void  MipsBiosAreaMemRead(_u8 * Mem, _u128 * fVal  , _u8 fDataType , _u64 fBase , int fOffset ) {

	_u64 RealAdd;
	RealAdd = ( fBase + fOffset ) & 0x3FFFFF;

	switch ( fDataType ) {



		case DATA_W : {
												MipsMemoryReadW (Mem , fVal , RealAdd );
												break;
										}

				
			case DATA_B : {
												MipsMemoryReadB ( Mem,  fVal , RealAdd );
												break;
										}

																	
			default: 	{
								printf("MemBIOS:Read DataType Not Handled\n");
								MipsCpu.halt = 1;
								break;
							}
		}

}



